class_name ShotSetup extends Node3D

enum Phase {
	AIM,
	POWER_ADJUST,
	CURVE_ADJUST,
	DOWNSWING,
	SHOT,
	FINISHED,
}

const PITCH_MIN := deg_to_rad(-60.0)
const PITCH_MAX := deg_to_rad(-5.0)
const ZOOM_LENGTH := 0.1
const ZOOM_ELEVATION_RATIO := 0.3166

const ZOOM_MIN := 1.0
const ZOOM_MAX := 12.0

const ARROW_ACCELERATION := 8.0
const PLAYER_ACCELERATION := 2.0

const FREE_CAM_RETURN_TIME := 0.618
const BALL_RETURN_TIME := 0.618
const CAMERA_SNAP_TIME := 0.3

## In Driving Range mode, the ball can be retrieved in the shot phase.
@export var driving_range := false

## Initially-selected club
@export var initial_club: Club.Type = Club.Type.DRIVER

@export_category("Shot Parameters")
@export var base_power := 2.5
@export var base_curve := 0.0

@export_category("Debug")
## When enabled, the game will pause and enter free cam mode when the ball has a collision.
@export var debug_ball_impact := false
## When enabled, ignore curve meter and hit a perfect shot every time.
@export var perfect_aim := false

var base_speed: float = ProjectSettings.get_setting("game/config/controls/camera/free_camera_speed")

var x_sensitivity: float = ProjectSettings.get_setting(
	"game/config/controls/camera/x_axis_sensitivity"
)
var x_acceleration: float = ProjectSettings.get_setting(
	"game/config/controls/camera/x_axis_acceleration"
)
var y_sensitivity: float = ProjectSettings.get_setting(
	"game/config/controls/camera/y_axis_sensitivity"
)
var y_acceleration: float = ProjectSettings.get_setting(
	"game/config/controls/camera/y_axis_acceleration"
)

var invert_pitch: bool = ProjectSettings.get_setting("game/config/controls/camera/invert_pitch")

var control_disabled := false

var phase: Phase:
	set(value):
		if value != phase:
			_on_phase_change(value)
		phase = value

var club: Club.Type:
	set(value):
		if value != club:
			_on_club_change(value)
		club = value

var shot_ref: Node3D

var _free_camera: FreeCamera
var _returning_free_camera := false

var _tracking_camera: OrbitalCamera

@onready var direction: Node3D = %Direction
@onready var elevation: Node3D = %Elevation
@onready var pitch: Node3D = %Pitch
@onready var zoom: Node3D = %Zoom
@onready var camera: Camera3D = %Camera

@onready var player_pivot: Node3D = %PlayerPivot
# TODO: genericize for selectable characters
@onready var character: CharacterController = $PlayerPivot/GfolfGirl

@onready var shot_animation: AnimationPlayer = %ShotAnimation

@onready var arrow: Node3D = %Arrow
@onready var arrow_pivot: Node3D = %ArrowPivot
@onready var arrow_animation: AnimationPlayer = %ArrowAnimation
@onready var shot_projection: ProjectileArc = %ShotProjection

@onready var power_bar: ProgressBar = %PowerBar
@onready var power_animation: AnimationPlayer = %PowerAnimation

@onready var curve_bar: ProgressBar = %CurveBar
@onready var curve_animation: AnimationPlayer = %CurveAnimation

@onready var ball_point: Node3D = %BallPoint
@onready var physics_ball: RigidBody3D = %PhysicsBall

@onready var drive_ref: RayCast3D = %DriveRef
@onready var drive_arrow: Node3D = %DriveArrow
@onready var wedge_ref: RayCast3D = %WedgeRef
@onready var wedge_arrow: Node3D = %WedgeArrow
@onready var iron_ref: RayCast3D = %IronRef
@onready var iron_arrow: Node3D = %IronArrow
@onready var putt_ref: RayCast3D = %PuttRef
@onready var putt_arrow: Node3D = %PuttArrow

@onready var club_selector: ClubSelector = %ClubSelector

@onready var nice_animation: AnimationPlayer = %NiceAnimation
@onready var wasted_animation: AnimationPlayer = %WastedAnimation

@onready var hud_state_machine: AnimationTree = %HUDStateMachine
@onready var hud_state: AnimationNodeStateMachinePlayback = hud_state_machine["parameters/playback"]

@onready var ball_impulse_debug: Node3D = %BallImpulseDebug

@onready var camera_distance := zoom.position.z:
	set = _set_camera_distance

@onready var phys_ball_scene := preload("res://src/player/physics_ball/physics_ball.tscn")

@onready var _target_rotation := Vector2(pitch.rotation.x, direction.rotation.y)


func _ready() -> void:
	_on_phase_change(phase)
	club = initial_club


func _set_camera_distance(value: float) -> void:
	var tween := get_tree().create_tween()
	tween.tween_property(zoom, "position:z", value, ZOOM_LENGTH).set_trans(Tween.TRANS_SINE)
	tween.set_parallel()
	(
		tween
		. tween_property(elevation, "position:y", value * ZOOM_ELEVATION_RATIO, ZOOM_LENGTH)
		. set_trans(Tween.TRANS_SINE)
	)
	camera_distance = value


func _unhandled_input(event: InputEvent) -> void:
	if event is InputEventMouseButton:
		Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)
	elif event is InputEventMouseMotion:
		if Input.get_mouse_mode() == Input.MOUSE_MODE_CAPTURED:
			camera_motion((event as InputEventMouseMotion).relative)


func camera_motion(motion: Vector2) -> void:
	if not control_disabled and phase == Phase.AIM:
		# Can only control camera while aiming
		_target_rotation.y = _target_rotation.y - deg_to_rad(motion.x * x_sensitivity)
		_target_rotation.x = clampf(
			_target_rotation.x - deg_to_rad(motion.y * y_sensitivity) * (-1 if invert_pitch else 1),
			PITCH_MIN,
			PITCH_MAX
		)


func finish_downswing() -> void:
	phase = Phase.SHOT


func get_shot_impulse(meter_pct: float) -> Vector3:
	return -shot_ref.global_basis.z * base_power * meter_pct


func take_shot() -> void:
	print_debug("WHACK!\nPower: ", power_bar.value, "\nCurve: ", curve_bar.value)

	var impulse := get_shot_impulse(power_bar.value)
	print_debug("Shot impulse: ", impulse, "; ", impulse.length(), " N*s")

	ball_impulse_debug.transform = (
		Transform3D.IDENTITY.scaled(Vector3.ONE * impulse.length()).looking_at(impulse)
	)

	physics_ball.freeze = false
	physics_ball.apply_central_impulse(impulse)


## Make the shot projection widget visible, with animated transition
func _show_shot_projection() -> void:
	shot_projection.initial_speed = 1
	shot_projection.basis = shot_ref.basis.orthonormalized()
	var shot_speed := get_shot_impulse(1.0).length() / physics_ball.mass
	var tween := get_tree().create_tween()
	tween.tween_property(shot_projection, "initial_speed", shot_speed, CAMERA_SNAP_TIME).set_trans(
		Tween.TRANS_QUAD
	)
	shot_projection.show()


func insert_free_cam() -> void:
	arrow_animation.play("hide")
	_show_shot_projection()
	hud_state.travel("hidden")
	_free_camera = FreeCamera.create(camera)
	add_sibling(_free_camera)
	control_disabled = true
	camera.current = false


func return_free_cam() -> void:
	# TODO alter shot aim based on free camera selection
	arrow_animation.play("show")
	shot_projection.hide()
	hud_state.travel("visible")
	_free_camera.queue_free()
	_free_camera = null
	control_disabled = false
	_returning_free_camera = false
	camera.current = true


func return_ball() -> void:
	physics_ball.freeze = true
	var tween := get_tree().create_tween()
	(
		tween
		. tween_property(
			physics_ball,
			"global_transform",
			ball_point.global_transform,
			BALL_RETURN_TIME,
		)
		. set_trans(Tween.TRANS_SINE)
	)
	tween.tween_callback(end_shot_track)


func travel_to_ball() -> void:
	physics_ball.freeze = true
	global_position = physics_ball.global_position
	physics_ball.global_transform = ball_point.global_transform


func start_shot_track() -> void:
	if phase == Phase.SHOT:
		_tracking_camera = OrbitalCamera.create(physics_ball)
		_tracking_camera.rotation.y = randf_range(0.0, TAU)
		add_sibling(_tracking_camera)
		_tracking_camera.global_transform = ball_point.global_transform


func end_shot_track() -> void:
	camera.make_current()
	if is_instance_valid(_tracking_camera):
		_tracking_camera.queue_free()
	if phase == Phase.SHOT:
		phase = Phase.FINISHED


## Called immediately before `club` is mutated.
func _on_club_change(new_club: Club.Type) -> void:
	# Hide all arrows
	# TODO animate?
	drive_arrow.hide()
	wedge_arrow.hide()
	iron_arrow.hide()
	putt_arrow.hide()
	club_selector.value = new_club
	match new_club:
		Club.Type.DRIVER:
			shot_ref = drive_ref
			drive_arrow.show()
			# TODO get this from player loadout
			# TODO club change animation
			character.hold_right(Club.catalog.debug_driver.get_model())
		Club.Type.PUTTER:
			shot_ref = putt_ref
			putt_arrow.show()
			character.hold_right(Club.catalog.debug_putter.get_model())
		Club.Type.WEDGE:
			# TODO
			shot_ref = wedge_ref
			wedge_arrow.show()
			character.hold_right(Club.catalog.debug_wedge.get_model())
		Club.Type.IRON:
			# TODO
			shot_ref = iron_ref
			iron_arrow.show()
			character.hold_right(Club.catalog.debug_iron.get_model())
		Club.Type.SPECIAL:
			# TODO
			print_debug("TODO: implement special clubs!")
		_:
			print_debug("Not sure how to equip club type: ", new_club)


## Called immediately before `phase` is mutated.
func _on_phase_change(new_phase: Phase) -> void:
	match new_phase:
		Phase.AIM:
			hud_state.travel("visible")
			if not arrow.visible:
				arrow_animation.play("show")
			camera.make_current()
			power_bar.hide()
			curve_bar.hide()

			character.reset()
		Phase.POWER_ADJUST:
			curve_bar.hide()

			power_bar.show()
			power_animation.stop()  # Reset if needed
		Phase.CURVE_ADJUST:
			curve_bar.show()
			curve_animation.play("fill")
		Phase.DOWNSWING:
			power_bar.hide()
			curve_bar.hide()
			character.downswing()
			shot_animation.play("swing_delay")  # calls `take_shot`
		Phase.SHOT:
			power_bar.hide()
			curve_bar.hide()

			nice_animation.play("display")

			if not driving_range:
				shot_animation.play("shoot")
			arrow_animation.play("hide")
			hud_state.travel("hidden")

			take_shot()
		Phase.FINISHED:
			power_bar.hide()
			curve_bar.hide()
			travel_to_ball()


func _process(delta: float) -> void:
	## Visual updates
	# Rotation
	direction.rotation.y = lerp_angle(
		direction.rotation.y, _target_rotation.y, delta * x_acceleration
	)
	pitch.rotation.x = lerp_angle(pitch.rotation.x, _target_rotation.x, delta * y_acceleration)

	# Arrow lags behind camera control
	arrow_pivot.rotation.y = lerp_angle(
		arrow_pivot.rotation.y, _target_rotation.y, delta * ARROW_ACCELERATION
	)
	# Player lags further behind
	player_pivot.rotation.y = lerp_angle(
		player_pivot.rotation.y, _target_rotation.y, delta * PLAYER_ACCELERATION
	)

	## Input Handling
	if control_disabled:
		if Input.is_action_just_pressed("camera_cancel"):
			if is_instance_valid(_free_camera) and not _returning_free_camera:
				_returning_free_camera = true
				var tween := get_tree().create_tween()
				(
					tween
					. tween_property(
						_free_camera,
						"global_transform",
						camera.global_transform,
						FREE_CAM_RETURN_TIME
					)
					. set_trans(Tween.TRANS_SINE)
				)
				tween.tween_callback(return_free_cam)
		return

	match phase:
		Phase.AIM:
			# Camera zoom
			if Input.is_action_just_pressed("shot_zoom_in"):
				camera_distance = max(camera_distance - 1.0, ZOOM_MIN)
			if Input.is_action_just_pressed("shot_zoom_out"):
				camera_distance = min(camera_distance + 1.0, ZOOM_MAX)

			# Club select
			if Input.is_action_just_pressed("select_driver"):
				club = Club.Type.DRIVER
			if Input.is_action_just_pressed("select_iron"):
				club = Club.Type.IRON
			if Input.is_action_just_pressed("select_wedge"):
				club = Club.Type.WEDGE
			if Input.is_action_just_pressed("select_special"):
				club = Club.Type.SPECIAL
			if Input.is_action_just_pressed("select_putter"):
				club = Club.Type.PUTTER

			# Switch to free cam
			if (
				Input.is_action_just_pressed("camera_back")
				or Input.is_action_just_pressed("camera_forward")
				or Input.is_action_just_pressed("camera_left")
				or Input.is_action_just_pressed("camera_right")
			):
				insert_free_cam()

			# Advance to next phase
			if Input.is_action_just_pressed("shot_accept"):
				phase = Phase.POWER_ADJUST
		Phase.POWER_ADJUST:
			if Input.is_action_just_pressed("shot_accept"):
				# TODO set power gauge parameters if needed
				character.start_upswing()
				power_animation.play("fill")
			if Input.is_action_just_pressed("ui_cancel"):
				power_animation.stop()
				phase = Phase.AIM
			if Input.is_action_just_released("shot_accept") and power_bar.value > 0:
				power_animation.pause()
				phase = Phase.CURVE_ADJUST
		Phase.CURVE_ADJUST:
			if Input.is_action_just_pressed("ui_cancel"):
				curve_animation.stop()
				phase = Phase.POWER_ADJUST
			if Input.is_action_just_pressed("shot_accept"):
				curve_animation.pause()
				phase = Phase.DOWNSWING
		Phase.SHOT:
			if driving_range and Input.is_action_just_pressed("shot_accept"):
				phase = Phase.AIM
				return_ball()
		Phase.FINISHED:
			# TODO: hand over to next player where appropriate
			phase = Phase.AIM


func _on_physics_ball_sleeping_state_changed() -> void:
	if physics_ball.sleeping and phase == Phase.SHOT:
		end_shot_track()


func _on_ball_entered_water() -> void:
	# Should only be possible during SHOT phase, but let's check just to be sure...
	if phase == Phase.SHOT:
		physics_ball.freeze = true
		wasted_animation.play("display")


func _on_physics_ball_body_entered(_body: Node) -> void:
	print_debug("BONK!")
	if debug_ball_impact:
		get_tree().paused = true
		var snap_point: Node3D = camera
		if _tracking_camera:
			snap_point = _tracking_camera
		_free_camera = FreeCamera.create(snap_point)
		add_sibling(_free_camera)
		control_disabled = true
		camera.current = false
